!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!    
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!    
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!    
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!module containing verious interpolation routines. See discription for more details !!!!!!
!D. Meerburg 2015!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

module interpolation 

use settings

implicit none 

contains

   subroutine spline (x, y, b, c, d, n)
!======================================================================
!  Calculate the coefficients b(i), c(i), and d(i), i=1,2,...,n
!  for cubic spline interpolation
!  s(x) = y(i) + b(i)*(x-x(i)) + c(i)*(x-x(i))**2 + d(i)*(x-x(i))**3
!  for  x(i) <= x <= x(i+1)
!  Alex G: January 2010
!----------------------------------------------------------------------
!  input..
!  x = the arrays of data abscissas (in strictly increasing order)
!  y = the arrays of data ordinates
!  n = size of the arrays xi() and yi() (n>=2)
!  output..
!  b, c, d  = arrays of spline coefficients
!  comments ...
!  spline.f90 program is based on fortran version of program spline.f
!  the accompanying function fspline can be used for interpolation
!======================================================================
implicit none
integer n
double precision x(n), y(n), b(n), c(n), d(n)
integer i, j, gap
double precision h

gap = n-1
! check input
if ( n < 2 ) return
if ( n < 3 ) then
  b(1) = (y(2)-y(1))/(x(2)-x(1))   ! linear interpolation
  c(1) = 0.
  d(1) = 0.
  b(2) = b(1)
  c(2) = 0.
  d(2) = 0.
  return
end if
!
! step 1: preparation
!
d(1) = x(2) - x(1)
c(2) = (y(2) - y(1))/d(1)
do i = 2, gap
  d(i) = x(i+1) - x(i)
  b(i) = 2.0*(d(i-1) + d(i))
  c(i+1) = (y(i+1) - y(i))/d(i)
  c(i) = c(i+1) - c(i)
end do
!
! step 2: end conditions 
!
b(1) = -d(1)
b(n) = -d(n-1)
c(1) = 0.0
c(n) = 0.0
if(n /= 3) then
  c(1) = c(3)/(x(4)-x(2)) - c(2)/(x(3)-x(1))
  c(n) = c(n-1)/(x(n)-x(n-2)) - c(n-2)/(x(n-1)-x(n-3))
  c(1) = c(1)*d(1)**2/(x(4)-x(1))
  c(n) = -c(n)*d(n-1)**2/(x(n)-x(n-3))
end if
!
! step 3: forward elimination 
!
do i = 2, n
  h = d(i-1)/b(i-1)
  b(i) = b(i) - h*d(i-1)
  c(i) = c(i) - h*c(i-1)
end do
!
! step 4: back substitution
!
c(n) = c(n)/b(n)
do j = 1, gap
  i = n-j
  c(i) = (c(i) - d(i)*c(i+1))/b(i)
end do
!
! step 5: compute spline coefficients
!
b(n) = (y(n) - y(gap))/d(gap) + d(gap)*(c(gap) + 2.0*c(n))
do i = 1, gap
  b(i) = (y(i+1) - y(i))/d(i) - d(i)*(c(i+1) + 2.0*c(i))
  d(i) = (c(i+1) - c(i))/d(i)
  c(i) = 3.*c(i)
end do
c(n) = 3.0*c(n)
d(n) = d(n-1)
end subroutine spline

  function ispline(u, x, y, b, c, d, n)
!======================================================================
! function ispline evaluates the cubic spline interpolation at point z
! ispline = y(i)+b(i)*(u-x(i))+c(i)*(u-x(i))**2+d(i)*(u-x(i))**3
! where  x(i) <= u <= x(i+1)
!----------------------------------------------------------------------
! input..
! u       = the abscissa at which the spline is to be evaluated
! x, y    = the arrays of given data points
! b, c, d = arrays of spline coefficients computed by spline
! n       = the number of data points
! output:
! ispline = interpolated value at point u
!=======================================================================
implicit none
double precision ispline
integer n
double precision  u, x(n), y(n), b(n), c(n), d(n)
integer i, j, k
double precision dx

! if u is ouside the x() interval take a boundary value (left or right)
if(u <= x(1)) then
  ispline = y(1)
  return
end if
if(u >= x(n)) then
  ispline = y(n)
  return
end if

!*
!  binary search for for i, such that x(i) <= u <= x(i+1)
!*
i = 1
j = n+1
do while (j > i+1)
  k = (i+j)/2
  if(u < x(k)) then
    j=k
    else
    i=k
   end if
end do
!*
!  evaluate spline interpolation
!*
dx = u - x(i)
ispline = y(i) + dx*(b(i) + dx*(c(i) + dx*d(i)))
end function ispline

      subroutine linint(x,y,xtab,ytab,ntab)
!c
!c     Given a value of x return a value of y based on interpolation
!c     within a table of y values (ytab) corresponding to the x values
!c     contained in the array xtab.  The subroutine assumes that the
!c     values in xtab increase monotonically.  Efficiency is increased
!c     by remembering the table points used in the last call (ilast)
!c     xtab and ytab are provided through the argument list as is their
!c     length ntab.
!c
!c    John Mahaffy 2/12/95
!c
      real(dl) :: xtab(ntab),ytab(ntab)
	  real(dl) :: x, y
	  integer :: ntab
	  real(dl) :: wx
	  integer :: i1, ilast
      save ilast

      data ilast/1/
!c
!c    Start the search from the last point of table use index
!c
      if (x.le.xtab(ilast+1)) then
!c
!c    Search down the table from point of last use
!c
          do 20 i1=ilast,1,-1
              if(x.ge.xtab(i1)) go to 60
  20          continue
          write(6,*) 'x = ', x, '  is below the table range'  
          stop
      else
!c
!c    Search up the table from point of last use
!c
          do 40 i1=ilast+1,ntab-1
              if(x.le.xtab(i1+1)) go to 60
  40          continue
          write(6,*) 'x = ', x, '  is above the table range'
          stop
      endif
!c
!c   Bounding points found, interpolate
!c
  60  wx=(x-xtab(i1))/(xtab(i1+1)-xtab(i1))
      y=(1-wx)*ytab(i1)+wx*ytab(i1+1)
      ilast=i1
      return
      end subroutine linint

      subroutine ThreeDlinear_interpolation(xmin,lintmax1,lintmax2,blll,l1,l2,l3,dx,y)
        !interpolation specifically for bispectrum with equal
        !vertices (x) and stepsizes (dx) and starting points (xmin)
        !and total number of steps (lintmax)
        integer, intent(in) :: lintmax1, lintmax2
        real(dl),intent(in) :: blll(lintmax1,lintmax2,lintmax2)
        integer, intent(in) :: xmin, dx
        real(dl), intent(in) :: l1, l2, l3
        real(dl), intent(out) :: y
        integer :: n1, n2, n3
        integer :: xz1,xz2,xz3
        real(dl) :: eL1, eL2, eL3
      
        !detrmine the cube to look for the value l1, l2, l3
        n1 = floor(1.d0*(l1-xmin)/dx)+1
        n2 = floor(1.d0*(l2-xmin)/dx)+1
        n3 = floor(1.d0*(l3-xmin)/dx)+1
        
        xz1 = lmin + (n1-1)*dx
        xz2 = lmin + (n2-1)*dx
        xz3 = lmin + (n3-1)*dx
        eL1 = (-xz1+l1)/real(dx)
        eL2 = (-xz2+l2)/real(dx)
        eL3 = (-xz3+l3)/real(dx)
        !write(*,*) l3, xz1, xz2, xz3, el1, el2, el3, n1, n2, n3
        y = blll(n1,n2,n3)*(1-eL1)*(1-eL2)*(1-eL3) + &
             blll(n1+1,n2,n3)*eL1*(1-eL2)*(1-eL3) + &
             blll(n1,n2+1,n3)*eL2*(1-eL1)*(1-eL3) + &
             blll(n1,n2,n3+1)*eL3*(1-eL2)*(1-eL1) + &
             blll(n1+1,n2,n3+1)*eL1*eL3*(1-eL2) + &
             blll(n1+1,n2+1,n3)*eL1*eL2*(1-eL3) + &
             blll(n1,n2+1,n3+1)*eL2*eL3*(1-eL1) + &
             blll(n1+1,n2+1,n3+1)*eL1*eL2*eL3
        
      end subroutine ThreeDlinear_interpolation

      
      subroutine ThreeDlinear_IrrGrid_interpolation(xmin,lintmax1,lintmax2,blll,l1,l2,l3,y)
        !interpolation specifically for bispectrum with equal
        !vertices (x) and stepsizes (dx) and starting points (xmin)
        !and total number of steps (lintmax)
        integer, intent(in) :: lintmax1, lintmax2
        real(dl),intent(in) :: blll(lintmax1,lintmax2,lintmax2)
        integer, intent(in) :: xmin
        real(dl), intent(in) :: l1, l2, l3
        real(dl), intent(out) :: y
        integer :: dx
        integer :: n1, n2, n3
        integer :: xz1,xz2,xz3
        real(dl) :: eL1, eL2, eL3
        !note that this is not very adaptive....
        !can make it more adaptive by making the ranges
        !below variables that can be associated with the ell_array
        if (l1 .le. 100) then
           dx  = gridDell(1)
           n1 = floor(1.d0*(l1-xmin)/dx)+1
           xz1 = xmin + (n1-1)*dx
           eL1 = (-xz1+l1)/real(dx)
        elseif (l1 .ge. 100 .and. l1 .le. 500) then
           dx  = gridDell(2)
           n1 = floor(1.d0*(l1-100)/dx)+1
           xz1 = 100 + (n1-1)*dx
           eL1 = (-xz1+l1)/real(dx)
           n1 = n1 + 90
        elseif (l1 .ge. 500) then 
           dx  = gridDell(3)
           n1 = floor(1.d0*(l1-500)/dx)+1
           xz1 = 500 + (n1-1)*dx
           eL1 = (-xz1+l1)/real(dx)
           n1 = n1 + 130
        endif
        if (l2 .le. 100) then
           dx  = gridDell(1)
           n2 = floor(1.d0*(l2-xmin)/dx)+1
           xz2 = xmin + (n2-1)*dx
           eL2 = (-xz2+l2)/real(dx)
        elseif (l2 .ge. 100 .and. l2 .le. 500) then 
           dx  = gridDell(2)
           n2 = floor(1.d0*(l2-100)/dx)+1
           xz2 = 100 + (n2-1)*dx
           eL2 = (-xz2+l2)/real(dx)
           n2 = n2 + 90
        elseif (l2 .ge. 500) then 
           dx  = gridDell(3)
           n2 = floor(1.d0*(l2-500)/dx)+1
           xz2 = 500 + (n2-1)*dx
           eL2 = (-xz2+l2)/real(dx)
           n2 = n2 + 130
        endif
        if (l3 .le. 100) then
           dx  = gridDell(1)
           n3 = floor(1.d0*(l3-xmin)/dx)+1
           xz3 = xmin + (n3-1)*dx
           eL3 = (-xz3+l3)/real(dx)
        elseif (l3 .ge. 100 .and. l3 .le. 500) then 
           dx  = gridDell(2)
           n3 = floor(1.d0*(l3-100)/dx)+1
           xz3 = 100 + (n3-1)*dx
           eL3 = (-xz3+l3)/real(dx)
           n3 = n3 + 90
        elseif (l3 .ge. 500) then 
           dx  = gridDell(3)
           n3 = floor(1.d0*(l3-500)/dx)+1
           xz3 = 500 + (n3-1)*dx
           eL3 = (-xz3+l3)/real(dx)
           n3 = n3 + 130
        endif
        !ugly, but ok for now
        if (n1 .ge. 332) n1 = 332
        if (n2 .ge. 332) n2 = 332
        if (n3 .ge. 332) n3 = 332
        !write(*,*) l3, xz1, xz2, xz3, el1, el2, el3, n1, n2, n3
        y = blll(n1,n2,n3)*(1-eL1)*(1-eL2)*(1-eL3) + &
             blll(n1+1,n2,n3)*eL1*(1-eL2)*(1-eL3) + &
             blll(n1,n2+1,n3)*eL2*(1-eL1)*(1-eL3) + &
             blll(n1,n2,n3+1)*eL3*(1-eL2)*(1-eL1) + &
             blll(n1+1,n2,n3+1)*eL1*eL3*(1-eL2) + &
             blll(n1+1,n2+1,n3)*eL1*eL2*(1-eL3) + &
             blll(n1,n2+1,n3+1)*eL2*eL3*(1-eL1) + &
             blll(n1+1,n2+1,n3+1)*eL1*eL2*eL3
        
      end subroutine ThreeDlinear_IrrGrid_interpolation
      
      subroutine OneDlinear_interpolation(xmin,lintmax,Cl,l1,dx,y)
        !interpolation specifically for bispectrum with equal
        !vertices (x) and stepsizes (dx) and starting points (xmin)
        !and total number of steps (lintmax)
        integer, intent(in) :: lintmax
        real(dl),intent(in) :: Cl(lintmax)
        integer, intent(in) :: xmin, dx
        real(dl), intent(in) :: l1
        real(dl), intent(out) :: y
        integer :: n1
        integer :: xz1
        real(dl) :: eL1
      
        !detrmine the interval to look for the value l1
        n1 = floor(1.d0*(l1-xmin)/dx)+1        
        xz1 = lmin + (n1-1)*dx
        eL1 = (-xz1+l1)/real(dx)
        !write(*,*) l3, xz1, xz2, xz3, el1, el2, el3, n1, n2, n3
	y = Cl(n1)*(1-eL1)+ Cl(n1+1)*eL1
	
      end subroutine OneDlinear_interpolation
      
      subroutine OneDlinear_IrrGrid_interpolation(xmin,lintmax,Cl,l1,y)
        !interpolation specifically for bispectrum with irragular
        !vertices (x) and stepsizes (dx) and starting points (xmin)
        !and total number of steps (lintmax)
        integer, intent(in) :: lintmax
        real(dl),intent(in) :: Cl(lintmax)
        integer, intent(in) :: xmin
        real(dl), intent(in) :: l1
        real(dl), intent(out) :: y
        integer :: dx
        integer :: n1
        integer :: xz1
        real(dl) :: eL1
      
        !detrmine the interval to look for the value l1
        if (l1 .le. 100) then
           dx  = gridDell(1)
           n1 = floor(1.d0*(l1-xmin)/dx)+1
           xz1 = xmin + (n1-1)*dx
           eL1 = (-xz1+l1)/real(dx)
           y = Cl(n1)*(1-eL1)+ Cl(n1+1)*eL1
        elseif (l1 .ge. 100 .and. l1 .le. 500) then 
           dx  = gridDell(2)
           n1 = floor(1.d0*(l1-100)/dx)+1
           xz1 = 100 + (n1-1)*dx
           eL1 = (-xz1+l1)/real(dx)
           n1  = n1 + 90
           y = Cl(n1)*(1-eL1)+ Cl(n1+1)*eL1
        elseif (l1 .ge. 500 .and.  l1 .le. 4500) then 
           dx  = gridDell(3)
           n1 = floor(1.d0*(l1-500)/dx)+1
           xz1 = 500 + (n1-1)*dx
           eL1 = (-xz1+l1)/real(dx)
           n1 = n1 + 130
           y = Cl(n1)*(1-eL1)+ Cl(n1+1)*eL1
        elseif (l1 .ge. 4500) then
           !max value...
           y = Cl(331)
        endif
           
!!$        !write(*,*) l3, xz1, xz2, xz3, el1, el2, el3, n1, n2, n3
        
        
      end subroutine OneDlinear_IrrGrid_interpolation
end module interpolation
